
SuperStrict

Rem
bbdoc: Graphics/GameUI
End Rem
Module Pub.GameUI

ModuleInfo "Version: 1.00"
ModuleInfo "Author: Starkkz"
ModuleInfo "License: Public Domain"

Import BRL.Blitz
Import BRL.Reflection
Import BRL.Graphics
Import BRL.Max2D
Import BRL.PNGLoader
Import BRL.Retro
Import BRL.Font
Import BRL.FreeTypeFont

Incbin "arial.ttf"
Incbin "pointers/default.png"
Incbin "pointers/link.png"
Incbin "pointers/textfield.png"
Incbin "close.png"
Incbin "minus.png"
Incbin "plus.png"
Incbin "next.png"
Incbin "prev.png"

Private
?Win32
Extern "Win32"
Function OpenClipboard:Int(hWnd:Int)
Function CloseClipboard:Int()
Function IsClipboardFormatAvailable:Int(Format:Int)
Function EmptyClipboard:Int()
Function GetClipboardData:Byte Ptr(Format:Int)
Function SetClipboardData:Int(Format:Int,hMem:Byte Ptr)
Function GlobalAlloc(Flags:Int, Bytes:Int)
Function GlobalFree(Mem:Int)
Function GlobalLock:Byte Ptr(Mem:Int)
Function GlobalUnlock(Mem:Int)
End Extern
?
Const CF_TEXT:Int = 1

Public

Rem
bbdoc: Base frame class
EndRem
Type TFrame
	Global DefaultFont:TImageFont = LoadImageFont("incbin::arial.ttf", 14)
	
	Field Name:String, Text:String
	Field x:Int, y:Int, w:Int, h:Int
	Field Font:TImageFont
	Field Pointer:TImage
	
	Field ColorR:Int = 255, ColorG:Int = 255, ColorB:Int = 255, ColorA:Float = 1
	
	Field Parent:TFrame, ID:Int
	Field Items:Object[], SelectedItem:Int = -1
	
	Field Child:TFrame[]
	Field ChildSorting:TList
	Field FocusedChild:TFrame
	
	Field Hidden:Int, Disabled:Int
	Field GrabX:Int, GrabY:Int
	
	Field Hint:String
	
	Method New()
		Self.Name = "Frame"
	EndMethod
	
	Method ParentCount:Int(Cond(Frame:TFrame)=Null)
		Local P:TFrame = Self.Parent
		Local Cnt:Int
		If Cond Then
			While P.Parent
				P = P.Parent
				If Cond(P) Then
					Cnt = Cnt + 1
				EndIf
			Wend
		Else
			While P.Parent
				P = P.Parent
				Cnt = Cnt + 1
			Wend
		EndIf
	EndMethod
	
	Method Empty:Int()
		If Not Self.ChildSorting Then Return True
		Return ListIsEmpty(Self.ChildSorting)
	EndMethod
	
	Method Alpha:Float()
		If Self.Parent Then
			Return Self.Parent.Alpha() * Self.ColorA
		EndIf
		Return Self.ColorA
	EndMethod
	
	Method getFont:TImageFont()
		If Self.Font Then Return Self.Font
		Return TFrame.DefaultFont
	EndMethod
	
	Method getX:Int()
		If Self.Parent Then
			Return Self.Parent.getX() + Self.x
		EndIf
		Return Self.x
	EndMethod
	
	Method getY:Int()
		If Self.Parent Then
			Return Self.Parent.getY() + Self.y
		EndIf
		Return Self.y
	EndMethod
	
	Method getWidth:Int()
		Return Self.w
	EndMethod
	
	Method getHeight:Int()
		Return Self.h
	EndMethod
	
	Method Focus()
		If Self.Parent Then
			Self.Parent.FocusChild(Self.ID)
		EndIf
	EndMethod
	
	Method Focused:Int()
		If Self.Parent Then
			Return Self.Parent.FocusedChild = Self
		EndIf
	EndMethod
	
	Method FocusChild(ChildID:Int)
		If ChildID > 0 Or ChildID < Self.Child.length Then
			Local ChildFrame:TFrame = Self.Child[ChildID]
			If ChildFrame Then
				If Self.ChildSorting Then
					ListRemove(Self.ChildSorting, ChildFrame)
					ListAddFirst(Self.ChildSorting, ChildFrame)
				EndIf
				Self.FocusedChild = ChildFrame
				Self.Focus()
			EndIf
		EndIf
	EndMethod
	
	Method SetShape(PosX:Int,PosY:Int,Width:Int,Height:Int)
		Self.x = PosX
		Self.y = PosY
		Self.w = Width
		Self.h = Height
	EndMethod
	
	Method SetHint(HintText:String)
		Self.Hint = HintText
	EndMethod
	
	Method AddFrame:Int(Frame:TFrame)
		Local ChildID:Int = Self.Child.length
		Self.Child = Self.Child[..ChildID+1]
		Self.Child[ChildID] = Frame
		Self.FocusedChild = Frame
		If Self.ChildSorting Then
			ListAddFirst(Self.ChildSorting, Frame)
		EndIf
		Frame.Parent = Self
		Frame.ID = ChildID
		Return ChildID
	EndMethod
	
	Method SetFrame:Int(Frame:TFrame,Index:Int)
		If Index < 0 Then Return -1
		If Index >= Self.Child.length Then
			Self.Child = Self.Child[..Index+1]
		EndIf
		Local ChildFrame:TFrame = Self.Child[Index]
		If ChildFrame Then
			RemoveFrame(ChildFrame)
		EndIf
		Self.Child[Index] = Frame
		Self.FocusedChild = Frame
		If Self.ChildSorting Then
			ListAddFirst(Self.ChildSorting, Frame)
		EndIf
		Frame.Parent = Self
		Frame.ID = Index
		Return Index
	EndMethod
	
	Method RemoveFrame(Frame:TFrame)
		If Frame.Parent = Self Then
			Local ChildID:Int = Frame.ID
			If ChildID = -1 Then
				For Local i:Int = 0 To Self.Child.length-1
					If Self.Child[i] = Frame Then
						ChildID = i
						Exit
					EndIf
				Next
			EndIf
			
			If ChildID >= 0 And ChildID < Self.Child.length
				Frame.Parent = Null
				Frame.ID = Null
				If ChildID = Self.Child.length-1 And ChildID > 0 Then
					Self.Child = Self.Child[..Self.Child.length - 1]
				Else
					Self.Child[ChildID] = Null
				EndIf
				If Self.FocusedChild = Frame Then
					Self.FocusedChild = Null
				EndIf
				If Self.ChildSorting Then
					ListRemove(Self.ChildSorting, Frame)
				EndIf
			EndIf
		EndIf
	EndMethod
	
	Method BeginRendering()
		SetViewport Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
		SetImageFont Self.getFont()
		SetColor Self.ColorR, Self.ColorG, Self.ColorB
		SetBlend ALPHABLEND
		SetAlpha Self.Alpha()
		SetScale 1, 1
	EndMethod
	
	Method RenderChild()
		If Self.ChildSorting Then
			For Local Frame:TFrame = EachIn Self.ChildSorting.Reversed()
				Frame.Render()
			Next
		EndIf
	EndMethod
	
	Method RenderHint()
		If Len(Self.Hint) > 0 Then
			If Self.CursorIdle() > 500 And Self.CurrentHover() = Self Then
				Local HintWidth:Int = TextWidth(Self.Hint)
				Local HintHeight:Int = TextHeight(Self.Hint)
				
				SetViewport Self.CursorX()-4, Self.CursorY()-4, HintWidth+8, HintHeight+8
				SetImageFont Self.getFont()
				SetAlpha 0.4
			
				SetColor 0, 0, 0
				DrawRect Self.CursorX()-4, Self.CursorY()-4, HintWidth+8, HintHeight+8
				
				SetColor 255, 255, 255
				DrawRect Self.CursorX()-3, Self.CursorY()-3, HintWidth+6, HintHeight+6
				
				SetColor 0, 0, 0
				DrawText Self.Hint, Self.CursorX(), Self.CursorY()
			EndIf
		EndIf
	EndMethod
	
	Method Render()
		Self.BeginRendering()
		Self.RenderChild()
		Self.RenderHint()
	EndMethod
	
	Method UpdateChild()
		For Local Frame:TFrame = EachIn Self.Child
			Frame.Update()
		Next
	EndMethod
	
	Method Update()
	EndMethod
	
	Method OnGrab()
	EndMethod
	
	Method OnDrop()
	EndMethod
	
	Method Grabbed:Int()
		Return Self.CurrentGrabbed() = Self
	EndMethod
	
	Method Show()
		Self.Hidden = False
		Self.Focus()
	EndMethod
	
	Method Hide()
		Self.Hidden = True
	EndMethod
	
	Method Enable()
		Self.Disabled = False
	EndMethod
	
	Method Disable()
		Self.Disabled = True
	EndMethod
	
	Method CursorX:Int()
		If Self.Parent Then
			Return Self.Parent.CursorX()
		EndIf
	EndMethod
	
	Method CursorY:Int()
		If Self.Parent Then
			Return Self.Parent.CursorY()
		EndIf
	EndMethod
	
	Method CursorZ:Int()
		If Self.Parent Then
			Return Self.Parent.CursorZ()
		EndIf
	EndMethod
	
	Method CursorIdle:Int()
		If Self.Parent Then
			Return Self.Parent.CursorIdle()
		EndIf
	EndMethod
	
	Method CursorHover:Int(PosX:Int,PosY:Int,Width:Int,Height:Int)
		Return Self.CursorX() >= Self.getX() + PosX And Self.CursorY() >= Self.getY() + PosY And Self.CursorX() <= Self.getX() + PosX + Width And Self.CursorY() <= Self.getY() + PosY + Height
	EndMethod
	
	Method CursorHoverArea:Int()
		Return Self.CursorHover(0, 0, Self.getWidth(), Self.getHeight()) And Not Self.Hidden
	EndMethod
	
	Method FirstFrame:TFrame()
		If Self.FocusedChild Then
			Return Self.FocusedChild.FirstFrame()
		EndIf
		Return Self
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.ChildSorting Then
			For Local ChildFrame:TFrame = EachIn Self.ChildSorting
				If ChildFrame.CursorHoverArea() Then
					Local HFrame:TFrame = ChildFrame.HoverFrame()
					If HFrame Then
						Return HFrame
					EndIf
				EndIf
			Next
		EndIf
		Return Self
	EndMethod
	
	Method CurrentHover:TFrame()
		If Self.Parent Then Return Self.Parent.CurrentHover()
	EndMethod
	
	Method CurrentFirst:TFrame()
		If Self.Parent Then Return Self.Parent.CurrentFirst()
	EndMethod
	
	Method CurrentGrabbed:TFrame()
		If Self.Parent Then Return Self.Parent.CurrentGrabbed()
	EndMethod
	
	Method Move(PosX:Int,PosY:Int)
		Self.x = PosX
		Self.y = PosY
	EndMethod
	
	Method SetText(Text:String)
		Self.Text = Text
	EndMethod
	
	Method Color(R:Int,G:Int,B:Int)
		Self.ColorR = R
		Self.ColorG = G
		Self.ColorB = B
	EndMethod
	
	Method Write(Text:String)
		If Self.FocusedChild Then
			Self.FocusedChild.Write(Text)
		EndIf
	EndMethod
	
	Method PasteText(Text:String)
		If Self.FocusedChild Then
			Self.FocusedChild.PasteText(Text)
		EndIf
	EndMethod
	
	Method CopiedText:String()
		If Self.FocusedChild Then
			Return Self.FocusedChild.CopiedText()
		EndIf
	EndMethod
	
	Method CutText:String()
		If Self.FocusedChild Then
			Return Self.FocusedChild.CutText()
		EndIf
	EndMethod
	
	Method AddItem:Int(Item:Object)
		Local Index:Int = Self.Items.length
		Self.Items = Self.Items[..Index+1]
		Self.Items[Index] = Item
		Return Index
	EndMethod
	
	Method SetItem(Index:Int,Item:Object)
		If Index >= 0 And Index < Self.Items.length Then
			Self.Items[Index] = Item
		ElseIf Index >= 0 And Index >= Self.Items.length Then
			Self.Items = Self.Items[..Index+1]
			Self.Items[Index] = Item
		EndIf
	EndMethod
	
	Method RemoveItem(Index:Int)
		If Index >= 0 And Index < Self.Items.length Then
			Self.Items[Index] = Null
			For Local i:Int = Self.Items.length-1 To 0 Step -1
				If Self.Items[i] = Null Then
					Self.Items = Self.Items[..i-1]
				Else
					Exit
				EndIf
			Next
		EndIf
	EndMethod
	
	Method getItemObject:Object(Index:Int)
		If Index >= 0 And Index < Self.Items.length
			Return Self.Items[Index]
		EndIf
	EndMethod
	
	Method SelectedItemObject:Object()
		Return Self.getItemObject(Self.SelectedItem)
	EndMethod
EndType

Rem
bbdoc: Render a frame
EndRem
Function RenderFrame(Frame:TFrame)
	Frame.Render()
EndFunction

Rem
bbdoc: Update a frame
EndRem
Function UpdateFrame(Frame:TFrame)
	Frame.Update()
EndFunction

Rem
bbdoc: Get the type name of a frame
returns: A string with the type name of the frame
EndRem
Function FrameName:String(Frame:TFrame)
	Return Frame.Name
EndFunction

Rem
bbdoc: Get the text of a frame
returns: A string that contains the text of a frame
EndRem
Function FrameText:String(Frame:TFrame)
	Return Frame.Text
EndFunction

Rem
bbdoc: To know wether a frame has been grabbed by an object
returns: True if grabbed, False otherwise
EndRem
Function FrameGrabbed:Int(Frame:TFrame)
	Return Frame.Grabbed()
EndFunction

Rem
bbdoc: Set the text of a frame
EndRem
Function SetFrameText(Frame:TFrame,Text:String)
	Frame.SetText(Text)
EndFunction

Rem
bbdoc: Set the color of a frame
EndRem
Function SetFrameColor(Frame:TFrame,R:Int,G:Int,B:Int)
	Frame.Color(R, G, B)
EndFunction

Rem
bbdoc: Changes the current selection of the frame item
EndRem
Function SelectFrameItem:Int(Frame:TFrame,Index:Int)
	If Index < Frame.Items.length Then
		Frame.SelectedItem = Index
		Return True
	EndIf
	Return False
EndFunction

Rem
bbdoc: The current selected frame item
EndRem
Function SelectedFrameItem:Int(Frame:TFrame)
	Return Frame.SelectedItem
EndFunction

Rem
bbdoc: Add an item to a frame, in this case the item is usually a string
returns: The index
EndRem
Function AddFrameItem:Int(Frame:TFrame,Item:Object)
	Return Frame.AddItem(Item)
EndFunction

Rem
bbdoc: Set an item to a frame in certain index
EndRem
Function SetFrameItem(Frame:TFrame,Index:Int,Item:Object)
	Frame.SetItem(Index, Item)
EndFunction

Rem
bbdoc: Remove an item from a frame in a certain index
EndRem
Function RemoveFrameItem(Frame:TFrame,Index:Int)
	Frame.RemoveItem(Index)
EndFunction

Rem
bbdoc: Set the hint of a frame
EndRem
Function SetFrameHint(Frame:TFrame,Hint:String)
	Frame.SetHint(Hint)
EndFunction

Rem
bbdoc: Get the current hint of a frame
EndRem
Function FrameHint:String(Frame:TFrame)
	Return Frame.Hint
EndFunction

Rem
bbdoc: Focus a frame, this is usually used for window frames
EndRem
Function FocusFrame(Frame:TFrame)
	Frame.Focus()
EndFunction

Rem
bbdoc: Set the position and size of a frame
EndRem
Function SetFrameShape(Frame:TFrame,x:Int,y:Int,w:Int,h:Int)
	Frame.SetShape(x, y, w, h)
EndFunction

Rem
bbdoc: Hide a frame
EndRem
Function HideFrame(Frame:TFrame)
	Frame.Hide()
EndFunction

Rem
bbdoc: Show a frame
EndRem
Function ShowFrame(Frame:TFrame)
	Frame.Show()
EndFunction

Rem
bbdoc: Disable a frame
EndRem
Function DisableFrame(Frame:TFrame)
	Frame.Disable()
EndFunction

Rem
bbdoc: Enable a frame
EndRem
Function EnableFrame(Frame:TFrame)
	Frame.Enable()
EndFunction

Rem
bbdoc: Get the current width of a frame
EndRem
Function FrameWidth:Int(Frame:TFrame)
	Return Frame.getWidth()
EndFunction

Rem
bbdoc: Get the current height of a frame
EndRem
Function FrameHeight:Int(Frame:TFrame)
	Return Frame.getHeight()
EndFunction

Rem
bbdoc: Know when a frame is disabled or not
returns: True if disabled, false otherwise
EndRem
Function FrameDisabled:Int(Frame:TFrame)
	Return Frame.Disabled
EndFunction

Rem
bbdoc: Know when a frame is hidden or not
returns: True if disabled, false otherwise
EndRem
Function FrameHidden:Int(Frame:TFrame)
	Return Frame.Hidden
EndFunction

Rem
bbdoc: Create a new button frame
EndRem
Function CreateButtonFrame:TFrame(Text:String,x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TButton = New TButton
	Frame.SetShape(x, y, w, h)
	Frame.SetText(Text)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Know when a button is grabbed or not
returns: True if grabbed, otherwise false
EndRem
Function ButtonStatus:Int(Frame:TFrame)
	Local Button:TButton = TButton(Frame)
	If Button Then
		Return Button.Grabbed()
	EndIf
EndFunction

Rem
bbdoc: Create a checkbox frame
EndRem
Function CreateCheckboxFrame:TFrame(Text:String,x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TCheckbox = New TCheckbox
	Frame.SetShape(x, y, w, h)
	Frame.SetText(Text)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Know when a checkbox is toggled or not
returns: True if toggled, false otherwise
EndRem
Function CheckboxStatus:Int(Frame:TFrame)
	Local CB:TCheckbox = TCheckbox(Frame)
	If CB Then
		Return CB.Status
	EndIf
EndFunction

Rem
bbdoc: Create a combobox
EndRem
Function CreateComboboxFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,MaxItems:Int=6,Separation:Int=15)
	Local Frame:TCombobox = New TCombobox
	Frame.SetShape(x, y, w, h)
	Frame.MaxItems = MaxItems
	Frame.Separation = Separation
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Change the current separation of a combobox
EndRem
Function SetComboboxSeparation(Frame:TFrame,Separation:Int=15)
	Local CB:TCombobox = TCombobox(Frame)
	If CB Then
		CB.SetSeparation(Separation)
	EndIf
EndFunction

Rem
bbdoc: Create a frame from certain graphics, this is the main frame used to create other frames
EndRem
Function CreateDesktopFrame:TFrame(Source:TGraphics)
	Local Frame:TDesktop = New TDesktop
	Frame.SourceGraphics = Source
	Return Frame
EndFunction

Rem
bbdoc: Create an image frame from certain image
EndRem
Function CreateImageFrame:TFrame(Image:TImage,x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TImageFrame = New TImageFrame
	Frame.SetShape(x, y, w, h)
	Frame.Image = Image
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Create a label frame
EndRem
Function CreateLabelFrame:TFrame(Text:String,x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TLabel = New TLabel
	Frame.SetShape(x, y, w, h)
	Frame.SetText(Text)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Create a listbox frame
EndRem
Function CreateListboxFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,Separation:Int=15)
	Local Frame:TListbox = New TListbox
	Frame.SetShape(x, y, w, h)
	Frame.SetSeparation(Separation)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Change the current separation of a listbox
EndRem
Function SetListboxSeparation(Frame:TFrame,Separation:Int=15)
	Local CB:TListbox = TListbox(Frame)
	If CB Then
		CB.SetSeparation(Separation)
	EndIf
EndFunction

Rem
bbdoc: A List View frame
EndRem
Function CreateListviewFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,Separation:Int=15)
	Local Frame:TListView = New TListView
	Frame.SetShape(x, y, w, h)
	Frame.SetSeparation(Separation)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Change the current separation of a ListView frame
EndRem
Function SetListviewSeparation(Frame:TFrame,Separation:Int=15)
	Local LV:TListView = TListView(Frame)
	If LV Then
		LV.SetSeparation(Separation)
	EndIf
EndFunction

Rem
bbdoc: Create a column for a ListView frame
EndRem
Function AddListviewColumn:TFrame(Source:TFrame,Text:String,Width:Int)
	Local LV:TListView = TListView(Source)
	If LV Then
		Local Frame:TFrame = New TFrame
		Frame.w = Width
		Frame.Text = Text
		LV.AddFrame(Frame)
		Return Frame
	EndIf
EndFunction

Rem
bbdoc: Create a column for a ListView frame in certain index
EndRem
Function SetListviewColumn:TFrame(Source:TFrame,Index:Int,Text:String,Width:Int)
	Local LV:TListView = TListView(Source)
	If LV Then
		Local ObjectFrame:Object = LV.GetItemObject(Index)
		Local Column:TFrame
		If ObjectFrame Then
			Column = TFrame(ObjectFrame)
		Else
			Column = New TFrame
			LV.SetItem(Index, Column)
		EndIf
		
		If Column Then
			Column.Text = Text
			Column.w = Width
			Return Column
		EndIf
	EndIf
EndFunction

Rem
bbdoc: Create a progress bar frame
EndRem
Function CreateProgressbarFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TProgbar = New TProgbar
	Frame.SetShape(x, y, w, h)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Change the current percent of a progress bar
EndRem
Function SetProgressbarPercent(Frame:TFrame,Percent:Float)
	Local Progbar:TProgbar = TProgbar(Frame)
	If Progbar Then
		Progbar.SetPercent(Percent)
	EndIf
EndFunction

Rem
bbdoc: Create a slider
EndRem
Function CreateSliderFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,Orientation:Int=SLIDER_VER)
	Local Frame:TSlider = New TSlider
	Frame.SetShape(x, y, w, h)
	Frame.Orientation = Orientation
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Change the current scale of a slider
EndRem
Function SetSliderScale(Frame:TFrame,ScaleA:Int,ScaleB:Int)
	Local Slider:TSlider = TSlider(Frame)
	If Slider Then
		Slider.SetSize(ScaleA, ScaleB)
	EndIf
EndFunction

Rem
bbdoc: Get the value of a slider
EndRem
Function GetSliderValue:Int(Frame:TFrame)
	Local Slider:TSlider = TSlider(Frame)
	If Slider Then
		Return Int(Slider.Value + 0.5)
	EndIf
EndFunction

Const SLIDER_HOR:Int = 1
Const SLIDER_VER:Int = 2

Rem
bbdoc: Create a tabber
EndRem
Function CreateTabberFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TTabber = New TTabber
	Frame.SetShape(x, y, w, h)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Create a text field
EndRem
Function CreateTextFieldFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TTextField = New TTextField
	Frame.SetShape(x, y, w, h)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Set a filter for a text field, the function must return 1 if it's not allowed to push the text and 0 if allowed.
EndRem
Function SetTextFieldFilter(Frame:TFrame,Filter:Int(Char:String))
	Local TextField:TTextField = TTextField(Frame)
	If TextField Then
		TextField.Filter = Filter
	EndIf
EndFunction

Rem
bbdoc: Get the begining of the selected area in a text field
returns: An integer representing the current position in a text field
EndRem
Function TextFieldSelectedStart:Int(Frame:TFrame)
	Local FieldFrame:TTextField = TTextField(Frame)
	If FieldFrame Then
		Return FieldFrame.SelectionStart()
	EndIf
EndFunction

Rem
bbdoc: Get the length of the selected area in a text field
returns: An integer representing the current length of the selected area in a text field
EndRem
Function TextFieldSelectedLen:Int(Frame:TFrame)
	Local FieldFrame:TTextField = TTextField(Frame)
	If FieldFrame Then
		Return Abs(FieldFrame.SelectionLength())
	EndIf
EndFunction

Rem
bbdoc: Enable the password mode in a text field
EndRem
Function EnableTextFieldPassword(Frame:TFrame)
	Local FieldFrame:TTextField = TTextField(Frame)
	If FieldFrame Then
		FieldFrame.Password = True
	EndIf
EndFunction

Rem
bbdoc: Disable the password mode in a text field
EndRem
Function DisableTextFieldPassword(Frame:TFrame)
	Local FieldFrame:TTextField = TTextField(Frame)
	If FieldFrame Then
		FieldFrame.Password = False
	EndIf
EndFunction

Rem
bbdoc: Create a TreeView
EndRem
Function CreateTreeViewFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,Separation:Int=15)
	Local Frame:TTreeView = New TTreeView
	Frame.SetShape(x, y, w, h)
	Frame.MainTreeViewNode.Separation = Separation
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Create a node in a TreeView
returns: A new frame object representing this node
EndRem
Function AddTreeViewNodeFrame:TFrame(Text:String,Source:TFrame,OpenImage:TImage=Null,CloseImage:TImage=Null,EmptyImage:TImage=Null,Separation:Int=15)
	If Source = Null Then Return Null
	If TTreeView(Source) Or TTreeViewNode(Source) Then
		Local Frame:TTreeViewNode = New TTreeViewNode
		Frame.SetText(Text)
		Frame.Separation = Separation
		Frame.OpenImage = OpenImage
		Frame.CloseImage = CloseImage
		Frame.EmptyImage = EmptyImage
		Source.AddFrame(Frame)
		Return Frame
	EndIf
EndFunction

Rem
bbdoc: Create a node in a TreeView at a certain index
returns: A new frame object representing this node
EndRem
Function SetTreeViewNodeFrame:TFrame(Text:String,Source:TFrame,Index:Int,OpenImage:TImage=Null,CloseImage:TImage=Null,EmptyImage:TImage=Null,Separation:Int=15)
	If Source = Null Then Return Null
	If TTreeView(Source) Or TTreeViewNode(Source) Then
		Local Frame:TTreeViewNode = New TTreeViewNode
		Frame.SetText(Text)
		Frame.Separation = Separation
		Frame.OpenImage = OpenImage
		Frame.CloseImage = CloseImage
		Frame.EmptyImage = EmptyImage
		Source.SetFrame(Frame, Index)
		Return Frame
	EndIf
EndFunction

Rem
bbdoc: Remove a node from a TreeView/TreeViewNode, note that this will not resort the index of the next nodes after this one
EndRem
Function FreeTreeViewNodeFrame(Frame:TFrame,Index:Int)
	If Frame And (Index >= 0 Or Index < Frame.Child.length) Then
		Local TreeView:TTreeView = TTreeView(Frame)
		If TreeView Then
			Local MainNode:TTreeViewNode = TreeView.MainTreeViewNode
			If MainNode Then
				Local Node:TTreeViewNode = TTreeViewNode(MainNode.Child[Index])
				If Node Then
					MainNode.RemoveFrame(Node)
				EndIf
			EndIf
		EndIf
	EndIf
EndFunction

Rem
bbdoc: Create a window
EndRem
Function CreateWindowFrame:TFrame(Title:String,x:Int,y:Int,w:Int,h:Int,Source:TFrame)
	Local Frame:TWindow = New TWindow
	Frame.SetShape(x, y, w, h)
	Frame.SetText(Title)
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Rem
bbdoc: Disable the close button of a window
EndRem
Function DisableWindowClose(Frame:TFrame)
	Local Window:TWindow = TWindow(Frame)
	If Window Then
		Window.Close = Null
	EndIf
EndFunction

Rem
bbdoc: Enable the close button of a window
EndRem
Function EnableWindowClose(Frame:TFrame)
	Local Window:TWindow = TWindow(Frame)
	If Window Then
		If Not Window.Close Then
			Window.Close = New TWindowClose
			Window.Close.Parent = Window
		EndIf
	EndIf
EndFunction

Rem
bbdoc: Create a canvas
EndRem
Function CreateCanvasFrame:TFrame(x:Int,y:Int,w:Int,h:Int,Source:TFrame,OnRender(Frame:TFrame))
	Local Frame:TCanvasFrame = New TCanvasFrame
	Frame.SetShape(x, y, w, h)
	Frame.OnRender = OnRender
	Source.AddFrame(Frame)
	Return Frame
EndFunction

Private
Function Copy(Text:String)
?Win32
	Local TextPtr:Byte Ptr = Text.ToCString()
	Local MemBlock:Int = GlobalAlloc($2|$2000, Text.length+1)
	Local DataPtr:Byte Ptr = GlobalLock(MemBlock)
	MemCopy DataPtr, TextPtr, Text.length
	GlobalUnlock(MemBlock)
	If OpenClipboard(0) Then
		EmptyClipboard
		SetClipboardData CF_TEXT, DataPtr
		CloseClipboard
	EndIf
?
EndFunction

Function Paste:String()
	Local Text:String
?Win32
	If OpenClipboard(0) Then
		Local TextPtr:Byte Ptr = GetClipboardData(CF_TEXT)
		Text = String.FromCString(TextPtr)
		CloseClipboard
	EndIf
?
	Return Text
EndFunction

Type TButton Extends TFrame
	Field HoverAlpha:Float

	Method New()
		Self.Name = "Button"
		Self.Pointer = LoadImage("incbin::pointers/link.png")
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
		EndIf
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor Self.ColorR*0.5, Self.ColorG*0.5, Self.ColorB*0.5
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()/2 - 1
			
			SetColor Self.ColorR*0.5 - 13, Self.ColorG*0.5 - 13, Self.ColorB*0.5 - 13
			DrawRect Self.getX()+1, Self.getY() + Self.getHeight()/2, Self.getWidth()-2, Self.getHeight()/2 - 1
			
			SetColor 255, 255, 255
			If Self.Grabbed() Then
				If Self.HoverAlpha < 0.3 Then
					Self.HoverAlpha = Self.HoverAlpha + 0.025
				ElseIf Self.HoverAlpha > 0.3 Then
					Self.HoverAlpha = Self.HoverAlpha - 0.025
				EndIf
			ElseIf Self.CurrentHover() = Self Then
				If Self.HoverAlpha < 0.5 Then
					Self.HoverAlpha = Self.HoverAlpha + 0.025
				EndIf
			ElseIf HoverAlpha > 0 Then
				Self.HoverAlpha = Self.HoverAlpha - 0.025
			EndIf
			
			If Self.HoverAlpha > 0 Then
				For Local i:Float = 1 To Self.getHeight()
					SetAlpha Self.Alpha() * (Self.getHeight() - i)/Self.getHeight() * Self.HoverAlpha
					DrawLine Self.getX()+1, Self.getY() + i, Self.getX() + Self.getWidth() - 2, Self.getY() + i
				Next
			EndIf
			
			SetAlpha Self.Alpha()
			DrawText Self.Text, Self.getX() - TextWidth(Self.Text)/2 + Self.getWidth()/2, Self.getY() - TextHeight(Self.Text)/2 + Self.getHeight()/2
			Self.RenderHint()
		EndIf
	EndMethod
EndType

Type TCheckbox Extends TFrame
	Field Status:Int

	Method New()
		Self.Name = "Checkbox"
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Status = Not Self.Status
		EndIf
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getHeight(), Self.getHeight()
			
			If Self.Status Then
				If Self.CurrentHover() = Self Then
					SetColor 255, 255, 255
				Else
					SetColor 210, 210, 210
				EndIf
				DrawRect Self.getX() + 1, Self.getY() + 1, Self.getHeight() - 2, Self.getHeight() - 2
			
				SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
				DrawRect Self.getX() + 3, Self.getY() + 3, Self.getHeight() - 6, Self.getHeight() - 6
			
				SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
				DrawRect Self.getX() + 4, Self.getY() + 4, Self.getHeight() - 8, Self.getHeight() - 8
			ElseIf Self.CurrentHover() = Self Then
				SetColor 255, 255, 255
				DrawRect Self.getX() + 1, Self.getY() + 1, Self.getHeight() - 2, Self.getHeight() - 2
			Else
				SetColor 210, 210, 210
				DrawRect Self.getX() + 1, Self.getY() + 1, Self.getHeight() - 2, Self.getHeight() - 2
			EndIf
			
			SetColor 0, 0, 0
			DrawText Self.Text, Self.getX() + Self.getHeight() + 2, Self.getY() - TextHeight(Self.Text)/2 + Self.getHeight()/2
			
			Self.RenderHint()
		EndIf
	EndMethod
EndType

Type TCombobox Extends TFrame
	Field Slider:TFrame
	Field Separation:Int
	Field MaxItems:Int
	Field Status:Int
	
	Field MinusImage:TImage = LoadImage("incbin::minus.png")
	Field PlusImage:TImage = LoadImage("incbin::plus.png")

	Method New()
		Self.Name = "Combobox"
		Self.Slider = CreateSliderFrame(0, 0, 15, 0, Self)
	EndMethod
	
	Method SetSeparation(Sep:Int=15)
		Self.Separation  = Sep
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.Slider.CursorHoverArea() Then
			Return Self.Slider
		EndIf
		Return Self
	EndMethod
	
	Method ItemText:String()
		Local CurrentItem:Object = Self.SelectedItemObject()
		Return String(CurrentItem)
	EndMethod
	
	Method CopiedText:String()
		Return Self.ItemText()
	EndMethod
	
	Method getHeight:Int()
		If Self.Status Then
			Return Self.h + Self.Separation * Self.MaxItems
		EndIf
		Return Self.h
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.w, Self.h
			
			SetColor 255, 255, 255
			DrawRect Self.getX()+1, Self.getY()+1, Self.w-2, Self.h-2
			
			SetViewport Self.getX() + 1, Self.getY() + 1, Self.w - 2, Self.h - 2
			SetColor 0, 0, 0
			DrawText Self.ItemText(), Self.getX() + 5, Self.getY() - TextHeight(Self.ItemText())/2 + Self.h/2
			
			SetColor 255, 255, 255
			If Self.Status Then
				DrawImageRect Self.MinusImage, Self.getX() + Self.getWidth() - Self.h + 1, Self.getY() + 1, Self.h - 2, Self.h - 2
			
				SetViewport Self.getX(), Self.getY() + Self.h, Self.getWidth(), Self.Separation * Self.MaxItems
				
				SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
				DrawRect Self.getX(), Self.getY() + Self.h, Self.getWidth(), Self.Separation * Self.MaxItems
				
				SetColor 255, 255, 255
				DrawRect Self.getX()+1, Self.getY() + Self.h + 1, Self.getWidth()-2, Self.Separation * Self.MaxItems - 2
				
				Local HoverItem:Float = Float(Self.CursorY() - Self.getY() - Self.h)/Self.Separation + Float(Self.Position())/Self.Separation
				
				SetViewport Self.getX() + 1, Self.getY() + Self.h + 1, Self.getWidth() - 2, Self.Separation * Self.MaxItems - 2
				SetColor 0, 0, 0
				For Local i:Int = Self.Position()/Self.Separation To Self.Position()/Self.Separation + Self.MaxItems
					If i >= 0 And i < Self.Items.length Then
						If Self.SelectedItem = i Then
							SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
							DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.h, Self.getWidth(), Self.Separation/2
							
							SetColor Self.ColorR*0.65, Self.ColorG*0.65, Self.ColorB*0.65
							DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2 + Self.h, Self.getWidth(), Self.Separation/2
							
							SetColor 0, 0, 0
						ElseIf Int(HoverItem) = i And Self.CurrentHover() = Self Then
							SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
							DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.h, Self.getWidth(), Self.Separation/2
							
							SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
							DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2 + Self.h, Self.getWidth(), Self.Separation/2
							
							SetColor 0, 0, 0
						EndIf
						DrawText String(Self.Items[i]), Self.getX()+2, Self.getY() + (i * Self.Separation) - Self.Position() + Self.h
					Else
						Exit
					EndIf
				Next
				
				RenderFrame(Self.Slider)
			Else
				DrawImageRect Self.PlusImage, Self.getX() + Self.getWidth() - Self.h + 1, Self.getY() + 1, Self.h - 2, Self.h - 2
			EndIf
			
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method Position:Int()
		Local SliderVal:Float = GetSliderValue(Self.Slider)
		Return SliderVal/(Self.Items.length * Self.Separation) * (Self.Items.length * Self.Separation - Self.MaxItems * Self.Separation)
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
			If Self.Status Then
				Local Selection:Float = Float(Self.GrabY - Self.h)/Self.Separation + Float(Self.Position())/Self.Separation
				Self.SelectedItem = Selection
			EndIf
			Self.Status = Not Self.Status
		EndIf
	EndMethod
	
	Method Update()
		If Self.Slider Then
			SetFrameShape(Self.Slider, Self.w - 15, Self.h + 1, 15, Self.Separation * Self.MaxItems - 2)
			Local SliderItems:Int = Self.Items.length - Self.MaxItems
			If SliderItems < 0 Then
				SliderItems = 0
			EndIf
			SetFrameColor(Self.Slider, Self.ColorR, Self.ColorG, Self.ColorB)
			SetSliderScale(Self.Slider, Self.MaxItems * Self.Separation, Self.Items.length * Self.Separation)
			If Not Self.Disabled Then
				UpdateFrame(Self.Slider)
			EndIf
		EndIf
	EndMethod
EndType

Type TDesktop Extends TFrame
	Field SourceGraphics:TGraphics, GraphicsHwnd:Int
	Field Cx:Int, Cy:Int, Cz:Int, CIdle:Int
	
	Field CurFirst:TFrame, CurHover:TFrame, CurGrab:TFrame
	Field LMD:Int[4], CMD:Int[4]
	
	Method New()
		Self.ChildSorting = CreateList()
		Self.Name = "Desktop"
		Self.Pointer = LoadImage("incbin::pointers/default.png")
	EndMethod
	
	Method CurrentFirst:TFrame()
		Return Self.CurFirst
	EndMethod
	
	Method CurrentHover:TFrame()
		Return Self.CurHover
	EndMethod
	
	Method CurrentGrabbed:TFrame()
		Return Self.CurGrab
	EndMethod
	
	Method CursorX:Int()
		Return Self.Cx
	EndMethod
	
	Method CursorY:Int()
		Return Self.Cy
	EndMethod
	
	Method CursorZ:Int()
		Return Self.Cz
	EndMethod
	
	Method CursorIdle:Int()
		Return MilliSecs() - Self.CIdle
	EndMethod
	
	Method RenderingPointer:TImage()
		If Self.CurHover Then
			If Self.CurHover.Pointer Then Return Self.CurHover.Pointer
		EndIf
		Return Self.Pointer
	EndMethod
	
	Method Render()
		SetGraphics Self.SourceGraphics
		Self.BeginRendering()
		Self.RenderChild()
		
		SetViewport Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
		Self.RenderHint()
		
		Self.BeginRendering()
		Local RPointer:TImage = Self.RenderingPointer()
		If RPointer Then
			HideMouse()
			If RPointer = Self.Pointer Then
				DrawImage RPointer, Self.CursorX(), Self.CursorY()
			Else
				DrawImage RPointer, Self.CursorX() - ImageWidth(RPointer)/2, Self.CursorY() - ImageHeight(RPointer)/2
			EndIf
		Else
			ShowMouse()
		EndIf
	EndMethod
	
	Method Update()
		SetGraphics Self.SourceGraphics
	
		Local Width:Int, Height:Int, Deph:Int, Hertz:Int, Flags:Int
		Self.SourceGraphics.GetSettings(Width, Height, Deph, Hertz, Flags)
		
		Self.CurFirst = Self.FirstFrame()
		Self.CurHover = Self.HoverFrame()
		
		Local Mx:Int = MouseX()
		Local My:Int = MouseY()
		
		If Self.Cx <> Mx Or Self.Cy <> My Then
			Self.CIDle = MilliSecs()
		EndIf
		
		Self.Cx = MouseX()
		Self.Cy = MouseY()
		Self.Cz = MouseZ()
		Self.w = Width
		Self.h = Height
		
		For Local i:Int = 1 To 3
			Self.LMD[i] = Self.CMD[i]
			Self.CMD[i] = MouseDown(i)
		Next
		
		If Not Self.LMD[1] And Self.CMD[1] Then
			If Self.CurHover <> Self Then
				Self.CurHover.GrabX = Self.Cx - Self.CurHover.getX()
				Self.CurHover.GrabY = Self.Cy - Self.CurHover.getY()
				Self.CurHover.OnGrab()
				Self.CurGrab = Self.CurHover
			EndIf
		ElseIf Self.LMD[1] And Not Self.CMD[1] Then
			If Self.CurGrab And Self.CurGrab <> Self Then
				Self.CurGrab.OnDrop()
				Self.CurGrab = Null
			EndIf
		EndIf
		
		Local Char:Int = GetChar()
		If Char = 3 Then
			Local Copied:String = Self.CopiedText()
			If Len(Copied) > 0 Then
				Copy(Copied)
			EndIf
		ElseIf Char = 22 Then
			Local Pasted:String = Paste()
			If Len(Pasted) > 0 Then
				Self.PasteText(Pasted)
			EndIf
		ElseIf Char = 24 Then
			Local Cut:String = Self.CutText()
			If Len(Cut) > 0 Then
				Copy(Cut)
			EndIf
		ElseIf Char >= 32 And Char <= 162 Or Char = 8 Then
			Self.Write(Chr(Char))
		EndIf
	
		Self.UpdateChild()
	EndMethod
EndType

Type TImageFrame Extends TFrame
	Field Image:TImage
	
	Method New()
		Self.Name = "Image"
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor Self.ColorR, Self.ColorG, Self.ColorB
			DrawImageRect Self.Image, Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
		EndIf
	EndMethod
EndType

Type TLabel Extends TFrame
	Method New()
		Self.Name = "Label"
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			DrawText Self.Text, Self.getX(), Self.getY()
			
			Self.RenderHint()
		EndIf
	EndMethod
EndType

Type TListbox Extends TFrame
	Field Slider:TFrame
	Field Separation:Int

	Method New()
		Self.Name = "Listbox"
		Self.Slider = CreateSliderFrame(0, 0, 15, 0, Self)
	EndMethod
	
	Method SetSeparation(Sep:Int=15)
		Self.Separation  = Sep
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.Slider.CursorHoverArea() Then
			Return Self.Slider
		EndIf
		Return Self
	EndMethod
	
	Method MaxItems:Int()
		Return Self.getWidth()/Self.Separation
	EndMethod

	Method CopiedText:String()
		Local CurrentItem:Object = Self.SelectedItemObject()
		Return String(CurrentItem)
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetViewport Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			
			SetColor 255, 255, 255
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			
			Local HoverItem:Float = Float(Self.CursorY() - Self.getY())/Self.Separation + Float(Self.Position())/Self.Separation
			
			SetColor 0, 0, 0
			For Local i:Int = Self.Position()/Self.Separation To Self.Position()/Self.Separation + Self.MaxItems()
				If i >= 0 And i < Self.Items.length Then
					If Self.SelectedItem = i Then
						SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
						DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position(), Self.getWidth(), Self.Separation/2
						
						SetColor Self.ColorR*0.6, Self.ColorG*0.6, Self.ColorB*0.6
						DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2, Self.getWidth(), Self.Separation/2
						
						SetColor 0, 0, 0
					ElseIf Int(HoverItem) = i And Self.CurrentHover() = Self Then
						SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
						DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position(), Self.getWidth(), Self.Separation/2
						
						SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
						DrawRect Self.getX(), Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2, Self.getWidth(), Self.Separation/2
						
						SetColor 0, 0, 0
					EndIf
					DrawText String(Self.Items[i]), Self.getX()+2, Self.getY() + (i * Self.Separation) - Self.Position()
				Else
					Exit
				EndIf
			Next
			
			If Self.Slider Then
				RenderFrame(Self.Slider)
			EndIf
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method Position:Int()
		Local SliderVal:Float = GetSliderValue(Self.Slider)
		Return SliderVal/(Self.Items.length * Self.Separation) * (Self.Items.length * Self.Separation - Self.getHeight())
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
			Local Selection:Float = Float(Self.GrabY)/Self.Separation + Float(Self.Position())/Self.Separation
			Self.SelectedItem = Selection
		EndIf
	EndMethod
	
	Method Update()
		If Self.Slider Then
			SetFrameShape(Self.Slider, Self.getWidth() - 15, 1, 15, Self.getHeight() - 2)
			Local SliderItems:Int = Self.Items.length - Self.MaxItems()
			If SliderItems < 0 Then
				SliderItems = 0
			EndIf
			SetFrameColor(Self.Slider, Self.ColorR, Self.ColorG, Self.ColorB)
			SetSliderScale(Self.Slider, Self.MaxItems() * Self.Separation, Self.Items.length * Self.Separation)
			If Not Self.Disabled Then
				Self.Slider.Update()
			EndIf
		EndIf
	EndMethod
EndType

Type TListView Extends TFrame
	Field SliderVert:TFrame
	Field SliderHor:TFrame
	Field Separation:Int
	
	Field ItemCount:Int
	Field ColumnsWidth:Int
	
	Method New()
		Self.Name = "Listview"
		Self.SliderVert = CreateSliderFrame(0, 0, 15, 0, Self)
		Self.SliderHor = CreateSliderFrame(0, 0, 0, 15, Self, SLIDER_HOR)
		Self.ChildSorting = CreateList()
	EndMethod
	
	Method AddFrame:Int(Frame:TFrame)
		Frame.Parent = Self
		If Frame.Name = "Frame" Then
			Super.AddFrame(Frame)
		EndIf
	EndMethod
	
	Method SetSeparation(Sep:Int=15)
		Self.Separation = Sep
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.SliderVert.CursorHoverArea() Then
			Return Self.SliderVert
		ElseIf Self.SliderHor.CursorHoverArea() Then
			Return Self.SliderHor
		EndIf
		Return Super.HoverFrame()
	EndMethod
	
	Method UpdateColumns:Int()
		Self.ItemCount = 0
		Self.ColumnsWidth = 0
		For Local Column:TFrame = EachIn Self.Child
			Local ColumnItems:Int = Column.Items.length
			If ColumnItems > Self.ItemCount Then
				Self.ItemCount = ColumnItems
			EndIf
			Self.ColumnsWidth = Self.ColumnsWidth + Column.w
		Next
	EndMethod
	
	Method MaxItems:Int()
		Return FrameHeight(Self.SliderVert) / Self.Separation
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor 255, 255, 255
			DrawRect Self.getX() + 1, Self.getY() + 22, Self.getWidth() - 2, Self.getHeight() - 23
			
			SetColor Self.ColorR, Self.ColorG, Self.ColorB
			DrawRect Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, 10
			
			SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
			DrawRect Self.getX() + 1, Self.getY() + 11, Self.getWidth() - 2, 10
			
			Local ColumnPosition:Int, HoverItem:Float = Float(Self.CursorY() - Self.getY() - 22)/Self.Separation + Float(Self.Position())/Self.Separation
			For Local Column:TFrame = EachIn Self.Child
				Local ColumnStart:Int = ColumnPosition + Self.getX() - Self.ColumnsPosition()
				Local ColumnWidth:Int = Column.w
				Local ColumnEnd:Int = ColumnStart + ColumnWidth
				Local RealColumnStart:Int = ColumnStart
				
				If ColumnStart < Self.getX() Then ColumnStart = Self.getX()
				If ColumnStart > Self.getX() + Self.getWidth() Then ColumnStart = Self.getX() + Self.getWidth()
				If ColumnEnd < Self.getX() Then ColumnEnd = Self.getX()
				If ColumnEnd > Self.getX() + Self.getWidth() Then ColumnEnd = Self.getX() + Self.getWidth()
				ColumnEnd = ColumnEnd - ColumnStart
				
				SetViewport ColumnStart, Self.getY(), ColumnEnd, Self.getHeight()
				SetColor 0, 0, 0
				DrawText Column.Text, RealColumnStart - TextWidth(Column.Text)/2 + ColumnWidth/2, Self.getY() - TextHeight(Column.Text)/2 + 10
				
				SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
				DrawLine RealColumnStart + ColumnWidth - 1, Self.getY() + 1, RealColumnStart + ColumnWidth - 1, Self.getY() + Self.getHeight() - 16
				
				SetViewport ColumnStart + 1, Self.getY() + 22, ColumnEnd - 1, Self.getHeight() - 22
				For Local i:Int = 0 To Column.Items.length - 1
					If Self.SelectedItem = i Then
						SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
						DrawRect RealColumnStart, Self.getY() + (i * Self.Separation) - Self.Position() + 22, Column.w - 1, Self.Separation/2
						
						SetColor Self.ColorR*0.6, Self.ColorG*0.6, Self.ColorB*0.6
						DrawRect RealColumnStart, Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2 + 22, Column.w - 1, Self.Separation/2
						
						SetColor 0, 0, 0
					ElseIf Int(HoverItem) = i And Self.CurrentHover() = Self Then
						SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
						DrawRect RealColumnStart, Self.getY() + (i * Self.Separation) - Self.Position() + 22, Column.w - 1, Self.Separation/2
						
						SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
						DrawRect RealColumnStart, Self.getY() + (i * Self.Separation) - Self.Position() + Self.Separation/2 + 22, Column.w - 1, Self.Separation/2
						
						SetColor 0, 0, 0
					EndIf
				
					SetColor 0, 0, 0
					DrawText String(Column.Items[i]), RealColumnStart + 5, Self.getY() + 22 + Self.Separation*i - Self.Position()
				Next
					
				ColumnPosition = ColumnPosition + ColumnWidth
			Next
			
			RenderFrame(Self.SliderVert)
			RenderFrame(Self.SliderHor)
		EndIf
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
			Local Selection:Float = Float(Self.GrabY - 22)/Self.Separation + Float(Self.Position())/Self.Separation
			Self.SelectedItem = Selection
		EndIf
	EndMethod
	
	Method Position:Int()
		Local SliderVal:Float = GetSliderValue(Self.SliderVert)
		Return SliderVal/(Self.ItemCount * Self.Separation) * (Self.ItemCount * Self.Separation - FrameHeight(Self.SliderVert))
	EndMethod
	
	Method ColumnsPosition:Int()
		Local SliderVal:Float = GetSliderValue(Self.SliderHor)
		Return SliderVal/Self.ColumnsWidth * (Self.ColumnsWidth - Self.getWidth())
	EndMethod
	
	Method Update()
		Self.UpdateColumns()
		Self.SliderVert.SetShape(Self.getWidth() - 15, 22, 15, Self.getHeight() - 37)
		Self.SliderHor.SetShape(0, Self.getHeight() - 15, Self.getWidth(), 15)
		
		SetSliderScale(Self.SliderVert, Self.MaxItems() * Self.Separation, Self.ItemCount * Self.Separation)
		SetSliderScale(Self.SliderHor, Self.getWidth(), Self.ColumnsWidth)
		
		If Not Self.Disabled And Not Self.Hidden Then
			UpdateFrame(Self.SliderVert)
			UpdateFrame(Self.SliderHor)
		EndIf
	EndMethod
EndType

Type TProgbar Extends TFrame
	Field Percent:Float
	
	Method SetPercent(Value:Float)
		Self.Percent = Value
	EndMethod

	Method New()
		Self.Name = "Progress Bar"
		Self.ColorR = 0
		Self.ColorB = 0
		Self.Percent = 1
	EndMethod

	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor 0, 0, 0
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor 255, 255, 255
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			
			SetColor Self.ColorR, Self.ColorG, Self.ColorB
			DrawRect Self.getX()+2, Self.getY()+2, Float(Self.getWidth()-4) * Self.Percent, Self.getHeight()-4
			
			SetColor 255, 255, 255
			For Local i:Float = 2 To Self.getHeight()/2
				SetAlpha Self.Alpha() * (Self.getHeight()/2 - i)/(Self.getHeight()/2)
				DrawLine Self.getX()+2, Self.getY() + i, Self.getX() + Self.getWidth() - 2, Self.getY() + i
			Next
			
			Local PercentInt:Int = Self.Percent*100
			Local PercentText:String = PercentInt+"%"
			SetColor 0, 0, 0
			SetAlpha Self.Alpha() * 1
			DrawText PercentText, Self.getX() + Self.getWidth()/2 - TextWidth(PercentText)/2, Self.getY() + Self.getHeight()/ 2 - TextHeight(PercentText)/2
			
			Self.RenderHint()
		EndIf
	EndMethod
EndType

Type TSlider Extends TFrame
	Field Orientation:Int
	Field SliderGrab:Int
	
	Field SizeA:Float
	Field SizeB:Float
	Field Value:Float

	Method New()
		Self.Name = "Slider"
	EndMethod
	
	Method SetSize(A:Int,B:Int)
		Self.SizeA = A
		Self.SizeB = B
		If Self.Value > Self.SizeB Then
			Self.Value = Self.SizeB
		ElseIf Self.Value < 0 Then
			Self.Value = 0
		EndIf
	EndMethod
	
	Method Percent:Float()
		If Self.SizeA > 0 And Self.SizeB > 0 And Self.SizeA < Self.SizeB Then
			Return Self.SizeA/Self.SizeB
		EndIf
		Return 1
	EndMethod
	
	Method bX:Float()
		If Self.SizeB > 0 And Self.Value > 0 Then
			Return (Self.Value/Self.SizeB * (Self.getWidth() - Self.bWidth() - 2)) + 1
		EndIf
		Return 1
	EndMethod
	
	Method bY:Float()
		If Self.SizeB > 0 And Self.Value > 0 Then
			Return (Self.Value/Self.SizeB * (Self.getHeight() - Self.bHeight() - 2)) + 1
		EndIf
		Return 1
	EndMethod
	
	Method bWidth:Float()
		Return Self.Percent() * Float(Self.getWidth()) - 2
	EndMethod
	
	Method bHeight:Float()
		Return Self.Percent() * Float(Self.getHeight()) - 2
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
			If Self.Orientation = SLIDER_VER Then
				Self.SliderGrab = Self.GrabY - Self.bY() + 1
				If Self.SliderGrab > Self.bHeight() Then Self.SliderGrab = Self.bHeight()/2
				If Self.SliderGrab < 0 Then Self.SliderGrab = Self.bHeight()/2
			ElseIf Self.Orientation = SLIDER_HOR Then
				Self.SliderGrab = Self.GrabX - Self.bX() + 1
				If Self.SliderGrab > Self.bWidth() Then Self.SliderGrab = Self.bWidth()/2
				If Self.SliderGrab < 0 Then Self.SliderGrab = Self.bWidth()/2
			EndIf
		EndIf
	EndMethod
	
	Method Update()
		If Not Self.Disabled And Not Self.Hidden Then
			If Self.Grabbed() Then
				Local Cursor:Float
				If Self.Orientation = SLIDER_VER Then
					Cursor = Self.CursorY() - Self.getY() - Self.SliderGrab
					Self.Value = Cursor/(Self.getHeight()-Self.bHeight()) * Self.SizeB
				ElseIf Self.Orientation = SLIDER_HOR Then
					Cursor = Self.CursorX() - Self.getX() - Self.SliderGrab
					Self.Value = Cursor/(Self.getWidth()-Self.bWidth()) * Self.SizeB
				EndIf
				
				If Self.Value > Self.SizeB Then Self.Value = Self.SizeB
				If Self.Value < 0 Then Self.Value = 0
			EndIf
		EndIf
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.6, Self.ColorG*0.6, Self.ColorB*0.6
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			If Self.Orientation = SLIDER_VER Then
				DrawRect Self.getX()+1, Self.getY() + Self.bY(), Self.getWidth()-2, Self.bHeight()
				
				If Self.Grabbed() Then
					SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
				ElseIf Self.CurrentHover() = Self Then
					SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
				Else
					SetColor Self.ColorR*0.6, Self.ColorG*0.6, Self.ColorB*0.6
				EndIf
				DrawRect Self.getX()+2, Self.getY() + Self.bY()+1, Self.getWidth()-4, Self.bHeight()-2
				
				SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
				DrawRect Self.getX() + Self.getWidth()/2 - Self.getWidth()/6, Self.getY() + Self.bY() + Self.bHeight()/2 - 2, Self.getWidth()/3, 1
				DrawRect Self.getX() + Self.getWidth()/2 - Self.getWidth()/6, Self.getY() + Self.bY() + Self.bHeight()/2, Self.getWidth()/3, 1
				DrawRect Self.getX() + Self.getWidth()/2 -Self.getWidth()/6, Self.getY() + Self.bY() + Self.bHeight()/2 + 2, Self.getWidth()/3, 1
			ElseIf Self.Orientation = SLIDER_HOR Then
				DrawRect Self.getX() + Self.bX(), Self.getY()+1, Self.bWidth(), Self.getHeight()-2
				
				If Self.Grabbed() Then
					SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
				ElseIf Self.CurrentHover() = Self Then
					SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
				Else
					SetColor Self.ColorR*0.6, Self.ColorG*0.6, Self.ColorB*0.6
				EndIf
				DrawRect Self.getX() + Self.bX()+1, Self.getY()+2, Self.bWidth()-2, Self.getHeight()-4
				
				SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
				DrawRect Self.getX() + Self.bX() + Self.bWidth()/2 - 2, Self.getY() + Self.getHeight()/2 - Self.getHeight()/6, 1, Self.getHeight()/3
				DrawRect Self.getX() + Self.bX() + Self.bWidth()/2, Self.getY() + Self.getHeight()/2 - Self.getHeight()/6, 1, Self.getHeight()/3
				DrawRect Self.getX() + Self.bX() + Self.bWidth()/2 + 2, Self.getY() + Self.getHeight()/2 - Self.getHeight()/6, 1, Self.getHeight()/3
			EndIf
			Self.RenderHint()
		EndIf
	EndMethod
EndType

Type TTabber Extends TFrame
	Field MoveNext:TTabberNext
	Field MovePrev:TTabberPrev
	Field Position:Int

	Method New()
		Self.Name = "Tabber"
		Self.MoveNext = New TTabberNext
		Self.MovePrev = New TTabberPrev
		Self.MoveNext.Parent = Self
		Self.MovePrev.Parent = Self
	EndMethod
	
	Method NextWidth:Int()
		If Not FrameHidden(Self.MoveNext) Then Return FrameWidth(Self.MoveNext)
	EndMethod
	
	Method PrevWidth:Int()
		If Not FrameHidden(Self.MovePrev) Then Return FrameWidth(Self.MovePrev)
	EndMethod
	
	Method Pos:Int()
		Return Self.Position - Self.PrevWidth()
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.ChildSorting Then
			For Local ChildFrame:TFrame = EachIn Self.ChildSorting
				If ChildFrame.CursorHoverArea() Then
					Return ChildFrame.HoverFrame()
				EndIf
			Next
		EndIf
		
		If Self.MoveNext And Not Self.MoveNext.Hidden And Self.MoveNext.CursorHoverArea() Then Return Self.MoveNext
		If Self.MovePrev And Not Self.MovePrev.Hidden And Self.MovePrev.CursorHoverArea() Then Return Self.MovePrev
		Return Self
	EndMethod

	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()

			Local CurrentW:Int
			For Local i:Int = 0 To Self.Items.length - 1
				Local ItemText:String = String(Self.Items[i])
				Local ItemW:Int = TextWidth(ItemText) + 5
			
				SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
				DrawRect Self.getX() - Self.Pos() + CurrentW, Self.getY(), ItemW, Self.getHeight()
				
				If Self.SelectedItem = i Then
					SetColor Self.ColorR*0.9, Self.ColorG*0.9, Self.ColorB*0.9
					DrawRect Self.getX() - Self.Pos() + CurrentW + 1, Self.getY() + 1, ItemW - 1 - (i = Self.Items.length - 1), Self.getHeight()/2
					
					SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
					DrawRect Self.getX() - Self.Pos() + CurrentW + 1, Self.getY() + Self.getHeight()/2 + 1, ItemW - 1 - (i = Self.Items.length - 1), Self.getHeight()/2 - Abs((Self.getHeight() Mod 2)*2 - 2) - (Self.getHeight() Mod 2)
				Else
					SetColor 255, 255, 255
					DrawRect Self.getX() - Self.Pos() + CurrentW + 1, Self.getY() + 1, ItemW - 1 - (i = Self.Items.length - 1), Self.getHeight() - 2
				EndIf
				
				SetColor 0, 0, 0
				DrawText ItemText, Self.getX() - Self.Pos() + CurrentW + (ItemW - TextWidth(ItemText))/2, Self.getY() - TextHeight(ItemText)/2 + Self.getHeight()/2
				
				CurrentW = CurrentW + ItemW
			Next
			
			RenderFrame(Self.MoveNext)
			RenderFrame(Self.MovePrev)
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled Then
			Local ItemsWidth:Int = -Self.Pos()
			For Local i:Int = 0 To Self.Items.length - 1
				Local ItemW:Int = TextWidth(String(Self.Items[i])) + 5
				If Self.GrabX >= ItemsWidth And Self.GrabY <= ItemsWidth + ItemW Then
					Self.SelectedItem = i
				EndIf
				ItemsWidth = ItemsWidth + ItemW
			Next
		EndIf
	EndMethod
	
	Method Update()
		If Not Self.Disabled Then
			Local ItemsWidth:Int
			For Local i:Int = 0 To Self.Items.length - 1
				ItemsWidth = ItemsWidth + TextWidth(String(Self.Items[i])) + 5
			Next
			
			If ItemsWidth < Self.getWidth() Then
				Self.MoveNext.Hidden = True
				Self.MovePrev.Hidden = True
			Else
				Self.MoveNext.Hidden = False
				Self.MovePrev.Hidden = False
			EndIf
		
			If Self.MoveNext.Grabbed() Then
				If Self.Position + Self.getWidth() < ItemsWidth + Self.NextWidth() + Self.PrevWidth() Then Self.Position = Self.Position + 4
				If Self.Position + Self.getWidth() > ItemsWidth + Self.NextWidth() + Self.PrevWidth() Then Self.Position = ItemsWidth - Self.getWidth() + Self.NextWidth() + Self.PrevWidth()
			ElseIf Self.MovePrev.Grabbed() Then
				If Self.Position > 0 Then Self.Position = Self.Position - 4
				If Self.Position < 0 Then Self.Position = 0
			EndIf
		EndIf
	EndMethod
EndType

Type TTabberNext Extends TFrame
	Field img:TImage = LoadImage("incbin::next.png")
	
	Method getX:Int()
		Return Self.Parent.getX() + Self.Parent.getWidth() - Self.getWidth()
	EndMethod
	
	Method getY:Int()
		Return Self.Parent.getY() - Self.getHeight()/2 + Self.Parent.getHeight()/2
	EndMethod
	
	Method getWidth:Int()
		Return ImageWidth(Self.img)
	EndMethod
	
	Method getHeight:Int()
		Return Self.Parent.getHeight()
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor 100, 100, 100
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			If Self.CurrentHover() = Self Then
				SetColor 255, 255, 255
			Else
				SetColor 200, 200, 200
			EndIf
			DrawRect Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
			DrawImageRect Self.img, Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
		EndIf
	EndMethod
EndType

Type TTabberPrev Extends TFrame
	Field img:TImage = LoadImage("incbin::prev.png")
	
	Method getX:Int()
		Return Self.Parent.getX()
	EndMethod
	
	Method getY:Int()
		Return Self.Parent.getY() - Self.getHeight()/2 + Self.Parent.getHeight()/2
	EndMethod
	
	Method getWidth:Int()
		Return ImageWidth(Self.img)
	EndMethod
	
	Method getHeight:Int()
		Return Self.Parent.getHeight()
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor 100, 100, 100
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			If Self.CurrentHover() = Self Then
				SetColor 255, 255, 255
			Else
				SetColor 200, 200, 200
			EndIf
			DrawRect Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
			DrawImageRect Self.img, Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
		EndIf
	EndMethod
EndType

Type TTextField Extends TFrame
	Field SelectStart:Int
	Field SelectEnd:Int
	Field Position:Int
	
	Field Password:Int
	
	Field Tick:Int, TimerTick:Int
	
	Field Filter:Int(Char:String)
	
	Method New()
		Self.Name = "Text Field"
		Self.Pointer = LoadImage("incbin::pointers/textfield.png")
	EndMethod
	
	Method PasteText(Text:String)
		Self.Write(Text)
	EndMethod
	
	Method CopiedText:String()
		Local Source:Int = Self.SelectionStart() + 1
		Local Length:Int = Abs(Self.SelectionLength())
		Return Mid(Self.Text, Source, Length)
	EndMethod
	
	Method CutText:String()
		Local Source:Int = Self.SelectionStart()
		Local Length:Int = Abs(Self.SelectionLength())
		If Length > 0 Then
			Local Data:String = Mid(Self.Text, Source, Length)
			Self.Write(Chr(8))
			Return Data
		EndIf
	EndMethod
	
	Method Write(Text:String)
		If Not Self.Disabled And Not Self.Hidden Then
			Local FixedText:String, OriginalText:String
			For Local i:Int = 0 To Text.length-1
				If Text[i] >= 32 And Text[i] <= 162 Or Text[i] = 8 Then
					If Text[i] = 8 Or Not Self.Filter Or Self.Filter(Chr(Text[i])) = 0 Then
						FixedText = FixedText + Chr(Text[i])
					EndIf
				EndIf
			Next
			OriginalText = Text
			Text = FixedText
		
			If Len(OriginalText) <= 0 Or Len(OriginalText) = Len(Text) Then
				Local LastPosition:Int = Self.GetPositionInText(Self.SelectStart)
				Local PartA:String = Mid(Self.Text, 1, Self.SelectionStart())
				Local PartB:String = Mid(Self.Text, Self.SelectionStart() + Abs(Self.SelectionLength()) + 1)
				If Text[0] = 8 And Len(Text) = 1 Then
					If Self.SelectionLength() <> 0 Then
						Self.SelectEnd = Self.SelectionStart()
						Self.SelectStart = Self.SelectEnd
						Self.Text = PartA + PartB
					Else
						Self.SelectEnd = Self.SelectionStart() - 1
						If Self.SelectEnd < 0 Then Self.SelectEnd = 0
						Self.SelectStart = Self.SelectEnd
						Self.Text = Mid(PartA, 1, Len(PartA) - 1) + PartB
					EndIf
				Else
					Self.SelectEnd = Self.SelectionStart() + Len(Text)
					Self.SelectStart = Self.SelectEnd
					Self.Text = PartA + Text + PartB
				EndIf
				
				Local NewPosition:Int = Self.GetPositionInText(Self.SelectStart)
				If Abs(NewPosition - Self.Position) > Self.getWidth() - 10 Then
					Self.Position = NewPosition - Self.getWidth() + 10
				ElseIf NewPosition < LastPosition And NewPosition > Self.getWidth() - 10 Then
					Self.Position = NewPosition - Self.getWidth() + 10
				ElseIf NewPosition < Self.getWidth() - 10 Then
					Self.Position = 0
				EndIf
			EndIf
		EndIf
	EndMethod
	
	Method GetPositionInText:Int(i:Int)
		If Self.Password Then
			Return TextWidth("*") * i
		EndIf
		Return TextWidth(Mid(Self.Text, 1, i))
	EndMethod
	
	Method GetMiddleInText:Int(Pos:Int)
		Pos = Pos + Self.Position
		Local Width:Int
		For Local i:Int = 1 To Len(Self.Text)
			Local Char:String = Mid(Self.Text, i, 1)
			Local CurrentWidth:Int
			If Self.Password Then
				CurrentWidth = TextWidth("*")
			Else
				CurrentWidth = TextWidth(Char)
			EndIf
			If Pos >= Width And Pos < Width + CurrentWidth Then
				Return i - 1
			EndIf
			Width = Width + CurrentWidth
		Next
		If Pos >= Width Then Return Len(Self.Text)
	EndMethod
	
	Method SelectionStart:Int()
		If Self.SelectEnd + 1 < Self.SelectStart Then
			Return Self.SelectEnd
		EndIf
		Return Self.SelectStart
	EndMethod
	
	Method DrawnText:String()
		Local Drawn:String = Self.Text
		If Self.Password Then
			Drawn = Null
			For Local i:Int = 1 To Len(Self.Text)
				Drawn = Drawn + "*"
			Next
		EndIf
		Return Drawn
	EndMethod
	
	Method DrawnTextWidth:Int()
		If Self.Password Then
			Return TextWidth("*") * Len(Self.Text)
		EndIf
		Return TextWidth(Self.Text)
	EndMethod
	
	Method DrawnTextHeight:Int()
		If Self.Password Then
			Return TextHeight("*")
		EndIf
		Return TextHeight(Self.Text)
	EndMethod
	
	Method SelectionLength:Int()
		Return Self.SelectEnd - Self.SelectStart
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
			Self.SelectStart = Self.GetMiddleInText(Self.CursorX() - Self.getX())
			Self.SelectEnd = Self.SelectStart
		EndIf
	EndMethod

	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetViewport Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			SetColor 255, 255, 255
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, Self.getHeight()-2
			
			SetColor 0, 0, 0
			DrawText Self.DrawnText(), Self.getX() - Self.Position + 5, Self.getY() - Self.DrawnTextHeight()/2 + Self.getHeight()/2
			
			If Self.SelectionLength() = 0 Then
				If Self.Tick Then
					If Self.CurrentFirst() = Self Then
						Local TickX:Int = Self.getX() + Self.GetPositionInText(Self.SelectStart) - Self.Position + 4
						DrawLine TickX, Self.getY() - Self.DrawnTextHeight()/2 + Self.getHeight()/2, TickX, Self.getY() + Self.DrawnTextHeight()/2 + Self.getHeight()/2
					EndIf
				EndIf
			ElseIf Self.SelectionLength() <> 0 Then
				If Self.CurrentFirst() = Self Then
					SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
				Else
					SetColor Self.ColorR*0.5, Self.ColorG*0.5, Self.ColorB*0.5
				EndIf
				SetAlpha Self.Alpha() * 0.7
				
				Local SelStart:Int = Self.GetPositionInText(Self.SelectStart)
				Local SelEnd:Int = Self.GetPositionInText(Self.SelectEnd)
				Local SelLen:Int = SelEnd - SelStart
				If SelStart < SelEnd Then
					DrawRect Self.getX() + SelStart - Self.Position + 5, Self.getY() - Self.DrawnTextHeight()/2 + Self.getHeight()/2, SelLen, Self.DrawnTextHeight()
				ElseIf SelEnd < SelStart Then
					DrawRect Self.getX() + SelEnd - Self.Position + 5, Self.getY() - Self.DrawnTextHeight()/2 + Self.getHeight()/2, -SelLen, Self.DrawnTextHeight()
				EndIf
			EndIf
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method Update()
		If Not Self.Disabled And Not Self.Hidden Then
			If Self.CurrentFirst() = Self Then
				If MilliSecs() - Self.TimerTick > 500 Then
					Self.TimerTick = MilliSecs()
					Self.Tick = Not Self.Tick
				EndIf
			ElseIf Tick Then
				Self.Tick = False
			EndIf
			
			If Self.Grabbed() Then
				Self.SelectEnd = Self.GetMiddleInText(Self.CursorX() - Self.getX())
				If Self.CursorX() - Self.getX() > Self.getWidth() And Self.Position < Self.DrawnTextWidth() - Self.getWidth() + 10 Then
					Self.Position = Self.Position + 2
				ElseIf Self.CursorX() - Self.getX() < 0 And Self.Position <> 0 Then
					Self.Position = Self.Position - 2
					If Self.Position < 0 Then
						Self.Position = 0
					EndIf
				EndIf
			EndIf
		EndIf
	EndMethod
EndType

Type TTreeView Extends TFrame
	Field SliderVert:TFrame
	Field SliderHor:TFrame
	Field MainTreeViewNode:TTreeViewNode
	
	Field TotalItems:Int
	Field TotalWidth:Int

	Method New()
		Self.ChildSorting = CreateList()
		Self.MainTreeViewNode = New TTreeViewNode
		Self.MainTreeViewNode.Parent = Self
		Self.MainTreeViewNode.Hidden = False
		Self.Name = "Treeview"
		
		Self.SliderVert = CreateSliderFrame(0, 0, 15, 0, Self)
		Self.SliderHor = CreateSliderFrame(0, 0, 0, 15, Self, SLIDER_HOR)
	EndMethod
	
	Method HoverFrame:TFrame()
		If SliderVert.CursorHoverArea() Then
			Return SliderVert
		ElseIf SliderHor.CursorHoverArea() Then
			Return SliderHor
		EndIf
		Return Self.MainTreeViewNode.HoverFrame()
	EndMethod
	
	Method AddFrame(Frame:TFrame)
		If TTreeViewNode(Frame) And Self.MainTreeViewNode Then
			Self.MainTreeViewNode.AddFrame(Frame)
		Else
			Super.AddFrame(Frame)
		EndIf
	EndMethod
	
	Method RemoveFrame(Frame:TFrame)
		Self.MainTreeViewNode.RemoveFrame(Frame)
	EndMethod
	
	Method SetFrame(Frame:TFrame,Index:Int)
		Self.MainTreeViewNode.SetFrame(Frame, Index)
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
		EndIf
	EndMethod
	
	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetColor Self.ColorR*0.4, Self.ColorG*0.4, Self.ColorB*0.4
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor 255, 255, 255
			DrawRect Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
			
			If Self.MainTreeViewNode Then
				SetViewport Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
				RenderFrame(Self.MainTreeViewNode)
			EndIf
			
			RenderFrame(Self.SliderVert)
			RenderFrame(Self.SliderHor)
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method HorPos:Int()
		Local SliderVal:Float = GetSliderValue(Self.SliderHor)
		Return SliderVal/Self.TotalWidth * (Self.TotalWidth - Self.getWidth() + 20)
	EndMethod
	
	Method VertPos:Int()
		Local SliderVal:Float = GetSliderValue(Self.SliderVert)
		Return SliderVal/(Self.TotalItems * Self.MainTreeViewNode.Separation) * (Self.TotalItems * Self.MainTreeViewNode.Separation - Self.getHeight() + 15)
	EndMethod
	
	Method Update()
		Self.SliderVert.SetShape(Self.getWidth() - 15, 0, 15, Self.getHeight() - 15)
		Self.SliderHor.SetShape(0, Self.getHeight() - 15, Self.getWidth(), 15)
		
		If Self.MainTreeViewNode Then
			Self.TotalItems = Self.MainTreeViewNode.TotalItems()
			Self.TotalWidth = Self.MainTreeViewNode.TotalWidth()
			
			SetSliderScale(Self.SliderVert, Self.getHeight(), Self.TotalItems * Self.MainTreeViewNode.Separation)
			SetSliderScale(Self.SliderHor, Self.getWidth(), Self.TotalWidth)
			
			If Not Self.Disabled And Not Self.Hidden Then
				If Self.SliderVert Then UpdateFrame(Self.SliderVert)
				If Self.SliderHor Then UpdateFrame(Self.SliderHor)
			EndIf
		
			If Not Self.Disabled And Not Self.Hidden Then
				If Self.MainTreeViewNode Then
					UpdateFrame(Self.MainTreeViewNode)
				EndIf
			EndIf
		EndIf
	EndMethod
EndType

Type TTreeViewNode Extends TFrame
	Field Separation:Int
	
	Field PlusImage:TImage = LoadImage("incbin::plus.png")
	Field MinusImage:TImage = LoadImage("incbin::minus.png")
	Field OpenImage:TImage
	Field CloseImage:TImage
	Field EmptyImage:TImage
	
	Field VarTotalWidth:Int
	Field VarTotalItems:Int

	Method New()
		Self.ChildSorting = CreateList()
		Self.Name = "Treeview Node"
		Self.Hidden = True
	EndMethod
	
	Method ParentSeparation:Int()
		Local TreeViewNodeParent:TTreeViewNode = TTreeViewNode(Self.Parent)
		Local TreeViewParent:TTreeView = TTreeView(Self.Parent)
		If TreeViewNodeParent Then
			Return TreeViewNodeParent.Separation
		ElseIf TreeViewParent Then
			Return Self.Separation
		EndIf
	EndMethod
	
	Method getWidth:Int()
		Return TextWidth(Self.Text) + 15
	EndMethod
	
	Method getHeight:Int()
		If Not Self.Hidden Then
			Return (Self.VarTotalItems + 1) * Self.ParentSeparation()
		EndIf
		Return Self.ParentSeparation()
	EndMethod
	
	Method getX:Int()
		Local TreeViewParent:TTreeView = TTreeView(Self.Parent)
		If TreeViewParent Then
			Return Super.getX() - TreeViewParent.HorPos() - 15
		EndIf
		Return Super.getX() + 15
	EndMethod
	
	Method ChildHeightBefore:Int(Index:Int)
		Local TreeViewNodeParent:TTreeViewNode = TTreeViewNode(Self.Parent)
		Local HeightBefore:Int
		For Local i:Int = 0 To Index - 1
			Local ChildNode:TTreeViewNode = TTreeViewNode(Self.Child[i])
			If ChildNode Then
				HeightBefore = HeightBefore + ChildNode.getHeight()
			EndIf
		Next
		Return Self.Separation + HeightBefore
	EndMethod
	
	Method getY:Int()
		Local TreeViewNodeParent:TTreeViewNode = TTreeViewNode(Self.Parent)
		If TreeViewNodeParent Then
			Return Super.getY() + TreeViewNodeParent.ChildHeightBefore(Self.ID)
		EndIf
		Local TreeViewParent:TTreeView = TTreeView(Self.Parent)
		If TreeViewParent Then
			Return Super.getY() - TreeViewParent.VertPos() - Self.ParentSeparation()
		EndIf
		Return Super.getY()
	EndMethod
	
	Method TotalWidth:Int()
		Local TotalWidth:Int = TextWidth(Self.Text) + 15
		If Not Self.Hidden Then
			For Local TreeViewChild:TFrame = EachIn Self.Child
				Local ChildNode:TTreeViewNode = TTreeViewNode(TreeViewChild)
				If ChildNode Then
					Local ChildTotalWidth:Int = ChildNode.TotalWidth()
					If ChildTotalWidth > TotalWidth Then
						TotalWidth = ChildTotalWidth
					EndIf
				EndIf
			Next
		EndIf
		Return TotalWidth
	EndMethod
	
	Method TotalItems:Int()
		Local ItemCount:Int
		If Not Self.Hidden Then
			ItemCount = Self.Child.length
			For Local TreeViewChild:TFrame = EachIn Self.Child
				Local ChildNode:TTreeViewNode = TTreeViewNode(TreeViewChild)
				If ChildNode Then
					ItemCount = ItemCount + ChildNode.TotalItems()
				EndIf
			Next
		EndIf
		Return ItemCount
	EndMethod
	
	Method Render()
		If TTreeViewNode(Self.Parent) Then
			SetColor 150, 150, 150
			If Parent.Child.length - 1 = Self.ID Then
				If Self.ID = 0 Then
					DrawLine Self.getX() + 7, Self.getY() + 7, Self.getX() + 7, Self.getY() + 7
				Else
					DrawLine Self.getX() + 7, Self.getY(), Self.getX() + 7, Self.getY() + 7
				EndIf
			Else
				If Self.ID = 0 Then
					DrawLine Self.getX() + 7, Self.getY() + 7, Self.getX() + 7, Self.getY() + Self.getHeight()
				Else
					DrawLine Self.getX() + 7, Self.getY(), Self.getX() + 7, Self.getY() + Self.getHeight()
				EndIf
			EndIf
			DrawLine Self.getX() + 7, Self.getY() + 7, Self.getX() + 12, Self.getY() + 7
			
			If Self.Parent.SelectedItem = Self.ID And Self.Child.length = 0 Then
				SetColor Self.ColorR*0.8, Self.ColorG*0.8, Self.ColorB*0.8
				DrawRect Self.getX() + 15, Self.getY(), TextWidth(Self.Text), TextHeight(Self.Text)
			EndIf
		
			SetColor 255, 255, 255
			If Not Self.Hidden Then
				If Self.CloseImage Then
					DrawImage Self.CloseImage, Self.getX(), Self.getY()
				ElseIf Self.Child.length > 0 Then
					DrawImage Self.MinusImage, Self.getX(), Self.getY()
				ElseIf Self.EmptyImage Then
					DrawImage Self.EmptyImage, Self.getX(), Self.getY()
				EndIf
			Else
				If Self.OpenImage Then
					DrawImage Self.OpenImage, Self.getX(), Self.getY()
				ElseIf Self.Child.length > 0 Then
					DrawImage Self.PlusImage, Self.getX(), Self.getY()
				ElseIf Self.EmptyImage Then
					DrawImage Self.EmptyImage, Self.getX(), Self.getY()
				EndIf
			EndIf
			
			SetColor 0, 0, 0
			DrawText Self.Text, Self.getX() + 15, Self.getY()
		EndIf
		
		If Not Self.Hidden Then
			Self.RenderChild()
		EndIf
		Self.RenderHint()
	EndMethod
	
	Method OnGrab()
		If Self.Child.length > 0 Then
			Self.Hidden = Not Self.Hidden
		EndIf
		Self.Focus()
		Self.Parent.SelectedItem = Self.ID
	EndMethod
	
	Method Update()
		Self.VarTotalWidth = Self.TotalWidth()
		Self.VarTotalItems = Self.TotalItems()
		Self.UpdateChild()
	EndMethod
EndType

Type TWindow Extends TFrame
	Field Closeable:Int
	Field Movable:Int = True
	
	Field Close:TWindowClose

	Method New()
		Self.ChildSorting = CreateList()
		Self.Name = "Window"
		Self.Close = New TWindowClose
		Self.Close.Parent = Self
	EndMethod
	
	Method CloseWidth:Int()
		If Self.Close Then
			Return FrameWidth(Self.Close)
		EndIf
	EndMethod
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
		EndIf
	EndMethod
	
	Method HoverFrame:TFrame()
		If Self.ChildSorting Then
			For Local ChildFrame:TFrame = EachIn Self.ChildSorting
				If ChildFrame.CursorHoverArea() Then
					Return ChildFrame.HoverFrame()
				EndIf
			Next
		EndIf
		If Self.Close And Self.Close.CursorHoverArea() Then
			Return Self.Close
		EndIf
		Return Self
	EndMethod

	Method Render()
		If Not Self.Hidden Then
			Self.BeginRendering()
			
			SetAlpha Self.Alpha()
			SetColor Self.ColorR*0.3, Self.ColorG*0.3, Self.ColorB*0.3
			DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
			
			SetColor Self.ColorR*0.7, Self.ColorG*0.7, Self.ColorB*0.7
			DrawRect Self.getX()+1, Self.getY()+20, Self.getWidth()-2, Self.getHeight()-21
			
			SetColor Self.ColorR*0.5, Self.ColorG*0.5, Self.ColorB*0.5
			DrawRect Self.getX()+1, Self.getY()+1, Self.getWidth()-2, 9
			
			SetColor Self.ColorR*0.5 - 13, Self.ColorG*0.5 - 13, Self.ColorB*0.45
			DrawRect Self.getX()+1, Self.getY()+10, Self.getWidth()-2, 9
			
			SetColor 255, 255, 255
			SetViewport Self.getX()+1, Self.getY()+1, Self.getWidth()-20, 19
			DrawText Self.Text, Self.getX() - TextWidth(Self.Text)/2 + Self.getWidth()/2 - Self.CloseWidth()/2, Self.getY() - TextHeight(Self.Text)/2 + 10
			
			If Self.Close Then
				RenderFrame(Self.Close)
			EndIf
			
			Self.RenderChild()
			Self.RenderHint()
		EndIf
	EndMethod
	
	Method Update()
		If Not Self.Disabled And Self.Movable Then
			If Self.Close Then
				Self.Close.SetShape(Self.getWidth() - 20, 3, 0, 0)
			EndIf
		
			If Self.Movable Then
				If Self.CurrentGrabbed() = Self And Self.GrabY <= 20 Then
					Local MoveX:Int = Self.CursorX() - Self.GrabX
					Local MoveY:Int = Self.CursorY() - Self.GrabY
					
					If MoveX < 0 Then MoveX = 0
					If MoveY < 0 Then MoveY = 0
					If Self.Parent Then
						If MoveX > Self.Parent.getWidth() - Self.getWidth() Then MoveX = Self.Parent.getWidth() - Self.getWidth()
						If MoveY > Self.Parent.getHeight() - Self.getHeight() Then MoveY = Self.Parent.getHeight() - Self.getHeight()
					EndIf
					Self.Move(MoveX, MoveY)
				EndIf
			EndIf
			Self.UpdateChild()
		EndIf
	EndMethod
EndType

Type TWindowClose Extends TFrame
	Field CloseImg:TImage = LoadImage("incbin::close.png")
	
	Method getWidth:Int()
		Return ImageWidth(Self.CloseImg)
	EndMethod
	
	Method getHeight:Int()
		Return ImageHeight(Self.CloseImg)
	EndMethod

	Method Render()
		Self.BeginRendering()
		If Self.Grabbed() Then
			SetColor 255*0.8, 255*0.8, 255*0.8
		ElseIf Self.CurrentHover() = Self Then
			SetColor 255, 255, 255
		Else
			SetColor 255*0.7, 255*0.7, 255*0.7
		EndIf
		DrawImage Self.CloseImg, Self.getX() - ImageWidth(Self.CloseImg)/2 + Self.getWidth()/2, Self.getY() - ImageHeight(Self.CloseImg)/2 + Self.getHeight()/2
	EndMethod
	
	Method OnGrab()
		HideFrame(Self.Parent)
	EndMethod
EndType

Type TCanvasFrame Extends TFrame
	Field OnRender(Frame:TFrame)
	
	Method OnGrab()
		If Not Self.Disabled And Not Self.Hidden Then
			Self.Focus()
		EndIf
	EndMethod

	Method Render()
		Self.BeginRendering()
		SetColor Self.ColorR*0.5, Self.ColorG*0.5, Self.ColorB*0.5
		DrawRect Self.getX(), Self.getY(), Self.getWidth(), Self.getHeight()
		
		SetColor 0, 0, 0
		DrawRect Self.getX() + 1, Self.getY() + 1, Self.getWidth() - 2, Self.getHeight() - 2
		
		Self.BeginRendering()
		Self.OnRender(Self)
	EndMethod
EndType